import json
import datetime as dt
import mbta.utils


class Response:

    """ Response
    
    Base class for responses from MBTA APIs
    
    """

    # In response type : (response columns) format
    column_map = {'travel_times': ('arr_dt', 'dep_dt', 'travel_time_sec', 'benchmark_travel_time_sec', 'direction',
                                   'route_id'
                                   )
                  }

    prettify_functions = dict()
    
    def __init__(self, raw_response, status_code):
        
        self.raw_response = json.loads(raw_response)
        self.data_as_of = dt.datetime.now()
        self.status_code = status_code
        self._set_data_type_data_list()

    @staticmethod
    def _strip_first_layer_of_dict(data):

        """ _strip_first_layer_of_dict

        MBTA responses come in a nested dictionary with one initial key. This will strip
        the first layer away to get a list of dictionaries.

        INPUTS

        data [dict]: Single dictionary. This will remove the first layer.


        RETURNS

        data_list [list of dicts]: List of subsequent data.

        """

        key = list(sorted(data.keys()))[0]

        data_list = data[key]

        return data_list, key

    @property
    def columns(self):

        """ columns
        column names in order for the response.

        """

        return self.column_map[self.data_type]

    def _set_data_type_data_list(self):

        """ _set_data_type_data_list

        Sets the data type and data list based on the raw response key and contents.

        """

        data_list_, data_type_ = self._strip_first_layer_of_dict(self.raw_response)

        self.data_list = data_list_
        self.data_type = data_type_

        return

    @property
    def tuples(self):

        """ tuples

        Data parsed into a list of tuples

        """

        tuples = []

        for data_point in self.data_list:

            # Sorts the each data point according to the column order in
            # self.columns, converts first to a list, then a tuple for faster processing later.
            tuple_ = tuple([data_point[key] for key in self.columns])
            tuples.append(tuple_)

        return tuples

    @property
    def pretty_tuples(self):

        """ pretty_tuples

        Data parsed into a list of tuples with prettify'd responses

        """

        tuples = []

        for data_point in self.data_list:
            # Sorts the each data point according to the column order in
            # self.columns, converts first to a list, then a tuple for faster processing later.
            tuple_ = tuple([self.prettify_response(key, data_point[key]) for key in self.columns])
            tuples.append(tuple_)

        return tuples

    def prettify_response(self, column_name, data_point):

        """ prettify_response

        Converts individual data points into something prettier, such as converting epoch timestamps to datetimes.

        INPUTS

        @column_name [str]: The column name to convert. Used to find the correct rule.

        @data_point [various]: The data point to convert.

        """

        transform_func = self.prettify_functions.get(column_name)

        if transform_func:
            return transform_func(data_point)

        return data_point


class MBTAPerformanceResponse(Response):
    
    """ MBTAPerformanceResponse
    
    Response class for the MBTA Performance API call results
    
    """

    prettify_functions = {
        'dep_dt': mbta.utils.epoch_to_datetime,
        'arr_dt': mbta.utils.epoch_to_datetime,
        'travel_time_sec': int,
        'benchmark_travel_time_sec': int,
    }
    
    def __init__(self, raw_response, status_code):

        super().__init__(raw_response=raw_response, status_code=status_code)



"""
filename: mbta/performance.py
author: Jared Stufft, jared@stufft.us
desc: Contains wrapper around the MBTA performance API. Allows a pythonic approach to retrieving historic performance
data for MBTA travels.
"""

import mbta.response
import mbta.utils


class MBTAPerformanceAPI:
    
    HOST = 'http://realtime.mbta.com/developer/api/v2.1'
    DOCUMENTATION = 'https://cdn.mbta.com/sites/default/files/developers/2018-10-30-mbta-realtime-performance-api' \
                    '-documentation-version-0-9-5-public.pdf'
    API_KEY_ENV_VARIABLE = 'MBTA_PERFORMANCE_API_KEY'

    def __init__(self, api_key=None):

        self.params = {
            'format': 'json',
            'api_key': mbta.utils.authorize_api(api_key, self.API_KEY_ENV_VARIABLE)
        }

    def get_travel_times(self, from_datetime, to_datetime, from_stop, to_stop, route=None):
        
        """ get_travel_times
        
        Retrieve travel time between two given stations for a given date range
        
        INPUTS
        
        @from_datetime [str]: a string in YYYY-MM-DD format denoting the beginning of the time interval to search for.

        @to_datetime [str]: a string in YYYY-MM-DD format denoting the end of the time interval to search for.

        @from_stop [str]: The stop_id for the beginning of the travel. Can be found using `get_gtfs_utility_data`
            function from mbta.utils with 'stops.txt' file.

        @to_stop [str]: The stop_id for the end of the travel. Can be found using `get_gtfs_utility_data`
            function from mbta.utils with 'stops.txt' file.

        @route [str]: The route name for the travel. If not included, will
        
        
        RETURNS
        
        @response [MBTAPerformanceResponse]: Response from the travel times API endpoint
        
        """

        params = {
            'from_datetime': mbta.utils.date_to_epoch(from_datetime),
            'to_datetime': mbta.utils.date_to_epoch(to_datetime),
            'from_stop': from_stop,
            'to_stop': to_stop
        }

        if route:
            params['route'] = route

        call_params = mbta.utils.merge_dicts(params, self.params)

        content, status_code = mbta.utils.make_api_call(self.HOST, ['traveltimes'], params=call_params)

        response = mbta.response.MBTAPerformanceResponse(content, status_code)
        
        return response




import setuptools

with open('README.md', 'r') as f:

    long_description = f.read()

setuptools.setup(
    name='mbta',
    version='0.0.4',
    author='Jared Stufft',
    author_email='jared@stufft.us',
    description='A simple, high-level Python wrapper for the MBTA API',
    long_description=long_description,
    url='https://github.com/JaredStufft/mbta',
    packages=setuptools.find_packages(),
    classifiers=(
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: Microsoft :: Windows',
        'Development Status :: 3 - Alpha'
    )
)


"""
filename: mbta/utils.py.py
author: Jared Stufft, jared@stufft.us
desc: Allows for access to some helper data and functions.
"""

import requests
import os
from zipfile import ZipFile
from io import BytesIO
import time


GTFS_CONTENT_URL = r'https://cdn.mbta.com/MBTA_GTFS.zip'


def get_gtfs_utility_data(file_name):

    """ get_gtfs_utility_data

    Retrieves the utility data for the GTFS network specific to the MBTA API. This contains things like stop_ids
    for other parts of the API. This content changes over time, so this function should allow the user to always
    have the latest distribution.

    INPUTS

    @file_name [str]: the file name for the utility data the user wishes to retrieve. The content comes as a zip
    file, so this chooses which piece of the zip to return.

    """

    r = requests.get(GTFS_CONTENT_URL)

    with ZipFile(BytesIO(r.content)) as zip_file:
        with zip_file.open(file_name) as stops_file:
            lines = stops_file.readlines()

    return lines


def date_to_epoch(date):

    """ date_to_epoch

    Takes a string in format YYYY-MM-DD and converts to epoch time

    INPUTS

    @date [str]: Date string in format YYYY-MM-DD


    RETURNS

    @epoch [int]: Integer value representing date in epoch format

    """

    epoch = int(time.mktime(time.strptime(date, '%Y-%m-%d')))

    return epoch


def epoch_to_datetime(epoch):

    """ _epoch_to_datetime

    Converts an integer epoch timestamp into a datetime string.

    INPUTS

    epoch [int]: Integer timestamp to convert to a datetime string.


    RETURNS

    date [str]: Converted datetime string in YYYy-MM-DD HH:MM:SS format.

    """

    date = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(int(epoch)))

    return date


def create_api_host_url(host, endpoints):

    """ create_api_host_url

    Creates the endpoint url for the API call.

    INPUTS

    @host [str]: The base URL for the API call.

    @endpoints [list]: The API endpoint names for the chosen method

    RETURNS

    @api_url [str]: The full endpoint url for the method.

    """

    api_url = '/'.join((host, *endpoints))

    return api_url


def merge_dicts(dict1, dict2):

    """ _merge_dicts

    Merges two dictionaries into one.

    INPUTS

    @dict1 [dict]: First dictionary to merge.

    @dict2 [dict]: Second dictionary to merge.


    RETURNS

    @merged [dict]: Merged dictionary

    """

    merged = {**dict1, **dict2}

    return merged


def make_api_call(host, endpoints, params):

    """ _make_api_call

    Base method for any API call.

    INPUTS

    @params [dict]: dictionary of API call parameters

    @endpoint [str]: the API method being used


    RETURNS

    @response [Response]: The response for the API call

    """

    call_url = create_api_host_url(host, endpoints)

    r = requests.get(call_url, params=params)

    r.raise_for_status()  # HTTPError if 4XX or 5XX status code on response.

    return r.content, r.status_code


def authorize_api(api_key, api_key_env):

    """ authorize_api

    Checks if the user provided an API key during class instance
    creation. If none found, searches for named environment variable.

    INPUTS

    @api_key [str]: Key for the particular API, to be received
    from the class instance created if used.


    RETURNS

    @valid_api_key [str]: The proper API key for the given API.

    """

    if not api_key:

        valid_api_key = os.getenv(api_key_env)

        return valid_api_key

    valid_api_key = api_key

    return valid_api_key
